unit main;

interface

uses
  Forms, Classes, Controls, ComCtrls, StdCtrls, ExtCtrls, ActnList;

type
  TfmMain = class(TForm)
    Panel: TPanel;
    TreeView: TTreeView;
    Splitter: TSplitter;
    ListView: TListView;
    btRefresh: TButton;
    ActionList: TActionList;
    acConnectable: TAction;
    acDiscoverable: TAction;
    btConnectable: TButton;
    btDiscoverable: TButton;
    acProperty: TAction;
    btProperty: TButton;
    btSelect: TButton;
    acEnable: TAction;
    btEnable: TButton;
    acRemove: TAction;
    btRemove: TButton;
    acUpdate: TAction;
    btUpdate: TButton;
    procedure FormCreate(Sender: TObject);
    procedure btRefreshClick(Sender: TObject);
    procedure TreeViewChange(Sender: TObject; Node: TTreeNode);
    procedure acConnectableUpdate(Sender: TObject);
    procedure acDiscoverableUpdate(Sender: TObject);
    procedure acConnectableExecute(Sender: TObject);
    procedure acDiscoverableExecute(Sender: TObject);
    procedure acPropertyUpdate(Sender: TObject);
    procedure acPropertyExecute(Sender: TObject);
    procedure btSelectClick(Sender: TObject);
    procedure acEnableUpdate(Sender: TObject);
    procedure acEnableExecute(Sender: TObject);
    procedure acRemoveUpdate(Sender: TObject);
    procedure acRemoveExecute(Sender: TObject);
    procedure acUpdateUpdate(Sender: TObject);
    procedure acUpdateExecute(Sender: TObject);
  end;

var
  fmMain: TfmMain = nil;

implementation

uses
  JwaBluetoothAPIs, Windows, SysUtils, Dialogs;

{$R *.dfm}

//      (dword)  ,
//      Bluetooth
function BTAdrToStr(const Adr: BLUETOOTH_ADDRESS): string;
var
  Loop: byte;
begin
  Result := IntToHex(Adr.rgBytes[0], 2);
  for Loop := 1 to 5 do
    Result := IntToHex(Adr.rgBytes[Loop], 2) + ':' + Result;
end;

procedure TfmMain.FormCreate(Sender: TObject);
begin
  btRefresh.Click;
end;

procedure TfmMain.btRefreshClick(Sender: TObject);
var
  RootNode: TTreeNode;
  hFind: HBLUETOOTH_RADIO_FIND;
  hDevFind: HBLUETOOTH_DEVICE_FIND;
  FindParams: BLUETOOTH_FIND_RADIO_PARAMS;
  SearchParams: BLUETOOTH_DEVICE_SEARCH_PARAMS;
  SearchParamsSize: dword;
  DevInfo: ^PBLUETOOTH_DEVICE_INFO;
  DevInfoSize: dword;
  hRadio: THandle;
  RadioInfo: PBLUETOOTH_RADIO_INFO;
  RadioInfoSize: dword;
  RadioNode: TTreeNode;
  Loop: integer;
  DevNode: TTreeNode;
begin
  with TreeView.Items do begin
    BeginUpdate;

    //  
    for Loop := 0 to Count - 1 do begin
      if TreeView.Items[Loop].ImageIndex > 0 then
        CloseHandle(TreeView.Items[Loop].ImageIndex);
      if Assigned(TreeView.Items[Loop].Data) then
        Dispose(TreeView.Items[Loop].Data);
    end;
    Clear;

    //    
    RootNode := Add(nil, 'Bluetooth Radios');
    with RootNode do begin
      Data := nil;
      ImageIndex := -1;
    end;

    //     Bluetooth
    FindParams.dwSize := SizeOf(BLUETOOTH_FIND_RADIO_PARAMS);
    hFind := BluetoothFindFirstRadio(@FindParams, hRadio);
    if hFind <> 0 then begin
      repeat
        //    
        New(RadioInfo);
        RadioInfoSize := SizeOf(BLUETOOTH_RADIO_INFO);
        FillChar(RadioInfo^, RadioInfoSize, 0);
        RadioInfo^.dwSize := RadioInfoSize;
        //   !!!
        BluetoothGetRadioInfo(hRadio, RadioInfo^);

        //    
        RadioNode := AddChild(RootNode,
                              string(RadioInfo^.szName) + ' [' +
                              BTAdrToStr(RadioInfo^.address) + ']');
        with RadioNode do begin
          //     Handle,    !
          ImageIndex := hRadio;
          Data := RadioInfo;
        end;

        //      .
        SearchParamsSize := SizeOf(BLUETOOTH_DEVICE_SEARCH_PARAMS);
        FillChar(SearchParams, SearchParamsSize, 0);
        SearchParams.dwSize := SearchParamsSize;
        SearchParams.fReturnRemembered := True;
        SearchParams.hRadio := hRadio;

        New(DevInfo);
        DevInfoSize := SizeOf(BLUETOOTH_DEVICE_INFO);
        FillChar(DevInfo^, DevInfoSize, 0);
        DevInfo^.dwSize := DevInfoSize;

        //  
        hDevFind := BluetoothFindFirstDevice(SearchParams, DevInfo^);
        if hDevFind <> 0 then begin
          repeat
            //   
            DevNode := AddChild(RadioNode,
                                string(DevInfo^.szName) + ' [' +
                                BTAdrToStr(DevInfo^.Address) + ']');
            with DevNode do begin
              Data := DevInfo;
              ImageIndex := -2;
            end;

            //   
            New(DevInfo);
            DevInfoSize := SizeOf(BLUETOOTH_DEVICE_INFO);
            FillChar(DevInfo^, DevInfoSize, 0);
            DevInfo^.dwSize := DevInfoSize;
          until not BluetoothFindNextDevice(hDevFind, DevInfo^);

          //   
          BluetoothFindDeviceClose(hDevFind);
        end;

        //   
      until not BluetoothFindNextRadio(hFind, hRadio);

      //   
      BluetoothFindRadioClose(hFind);
    end;

    EndUpdate;
  end;

  with TreeView do begin
    Selected := RootNode;
    Items[0].Expand(True);
  end;
end;

procedure TfmMain.TreeViewChange(Sender: TObject; Node: TTreeNode);
var
  ASelected: TTreeNode;

  procedure ShowRadios;
  var
    Info: PBLUETOOTH_RADIO_INFO;
    CurNode: TTreeNode;
  begin
    //  
    with ListView.Columns do begin
      BeginUpdate;
      with Add do Caption := 'Address';
      with Add do Caption := 'Name';
      with Add do Caption := 'Class Of Device';
      with Add do Caption := 'Manufacturer';
      with Add do Caption := 'Subversion';
      with Add do Caption := 'Connectable';
      with Add do Caption := 'Discoverable';
      EndUpdate;
    end;

    //  
    with ListView.Items do begin
      BeginUpdate;

      CurNode := ASelected.GetFirstChild;

      while Assigned(CurNode) do begin
        Info := PBLUETOOTH_RADIO_INFO(CurNode.Data);

        //    
        BluetoothGetRadioInfo(CurNode.ImageIndex, Info^);

        with Add do begin
          Data := Pointer(CurNode.ImageIndex);
          Caption := BTAdrToStr(Info.address);
          with SubItems do begin
            Add(string(Info.szName));
            Add(IntToStr(Info.ulClassofDevice));
            Add(IntToStr(Info.manufacturer));
            Add(IntToStr(Info.lmpSubversion));
            // NEW FUNCTIONS!!!
            Add(BoolToStr(BluetoothIsConnectable(CurNode.ImageIndex),
                          True));
            Add(BoolToStr(BluetoothIsDiscoverable(CurNode.ImageIndex),
                          True));
          end;
        end;

        CurNode := ASelected.GetNextChild(CurNode);
      end;

      EndUpdate;;
    end;
  end;

  procedure ShowDevices;
  var
    Info: ^PBLUETOOTH_DEVICE_INFO;
    CurNode: TTreeNode;
  begin
    //  
    with ListView.Columns do begin
      BeginUpdate;
      with Add do Caption := 'Address';
      with Add do Caption := 'Name';
      with Add do Caption := 'Class Of Device';
      with Add do Caption := 'Connected';
      with Add do Caption := 'Remembered';
      with Add do Caption := 'Authenticated';
      with Add do Caption := 'Last Seen';
      with Add do Caption := 'Last Used';
      EndUpdate;
    end;

    //  
    with ListView.Items do begin
      BeginUpdate;

      CurNode := ASelected.GetFirstChild;

      while Assigned(CurNode) do begin
        Info := CurNode.Data;

        //    
        //    ,   
        //     ,    
        BluetoothGetDeviceInfo(ASelected.ImageIndex, Info^);

        with Add do begin
          Data := Info;
          Caption := BTAdrToStr(Info^.Address);
          with SubItems do begin
            Add(string(Info^.szName));
            Add(IntToStr(Info^.ulClassofDevice));
            Add(BoolToStr(Info^.fConnected, True));
            Add(BoolToStr(Info^.fRemembered, True));
            Add(BoolToStr(Info^.fAuthenticated, True));
            try // stLastSeen   0     
              Add(DateTimeToStr(SystemTimeToDateTime(Info^.stLastSeen)));
            except
              Add('');
            end;
            try // stLastUsed   0     
              Add(DateTimeToStr(SystemTimeToDateTime(Info^.stLastUsed)));
            except
              Add('');
            end;
          end;
        end;

        CurNode := ASelected.GetNextChild(CurNode);
      end;

      EndUpdate;;
    end;
  end;

  procedure ShowServices;
  var
    Info: __PBLUETOOTH_DEVICE_INFO;
    ServiceCount: dword;
    Services: array of TGUID;
    hRadio: THandle;
    Loop: integer;
  begin
    //  
    with ListView.Columns do begin
      BeginUpdate;
      with Add do Caption := 'GUID';
      EndUpdate;
    end;

    //  
    with ListView.Items do begin
      BeginUpdate;

      //    
      ServiceCount := 0;
      Services := nil;
      hRadio := ASelected.Parent.ImageIndex;
      Info := ASelected.Data;
      // NEW FUNCTION
      BluetoothEnumerateInstalledServices(hRadio, Info, ServiceCount, nil);

      //  .
      SetLength(Services, ServiceCount);

      //   
      BluetoothEnumerateInstalledServices(hRadio, Info, ServiceCount,
                                          PGUID(Services));

      //  
      for Loop := 0 to ServiceCount - 1 do
        with Add do 
          Caption := GUIDToString(Services[Loop]);

      //  
      Services := nil;

      EndUpdate;
    end;
  end;

begin
  ASelected := TreeView.Selected;

  //  ListView
  with ListView do begin
    with Columns do begin
      BeginUpdate;
      Clear;
      EndUpdate;
    end;

    with Items do begin
      BeginUpdate;
      Clear;
      EndUpdate;
    end;
  end;

  //  
  if Assigned(ASelected) then
    case ASelected.ImageIndex of
      -2: ShowServices;
      -1: ShowRadios;
    else
      if ASelected.ImageIndex > 0 then ShowDevices;
    end;
end;

procedure TfmMain.acConnectableUpdate(Sender: TObject);
var
  SelectedItem: TListItem;
  SelectedNode: TTreeNode;
begin
  SelectedNode := TreeView.Selected;
  SelectedItem := ListView.Selected;

  with TAction(Sender) do begin
    Enabled := Assigned(SelectedNode) and Assigned(SelectedItem) and
               (SelectedNode.ImageIndex = -1);
               
    if Enabled then
      if StrToBool(SelectedItem.SubItems[4]) then
        Caption := 'Not conn.'
      else
        Caption := 'Connectable';
  end;
end;

procedure TfmMain.acDiscoverableUpdate(Sender: TObject);
var
  SelectedItem: TListItem;
  SelectedNode: TTreeNode;
begin
  SelectedNode := TreeView.Selected;
  SelectedItem := ListView.Selected;

  with TAction(Sender) do begin
    Enabled := Assigned(SelectedNode) and Assigned(SelectedItem) and
               (SelectedNode.ImageIndex = -1);

    if Enabled then
      if StrToBool(SelectedItem.SubItems[5]) then
        Caption := 'Not disc.'
      else
        Caption := 'Discoverable';
  end;
end;

procedure TfmMain.acConnectableExecute(Sender: TObject);
var
  SelectedItem: TListItem;
begin
  SelectedItem := ListView.Selected;

  if Assigned(SelectedItem) then
    if not BluetoothEnableIncomingConnections(Integer(SelectedItem.Data),
                  TAction(Sender).Caption = 'Not conn.') then
      MessageDlg('Unable to change Radio state', mtError, [mbOK], 0)

    else
      TreeViewChange(TreeView, TreeView.Selected);
end;

procedure TfmMain.acDiscoverableExecute(Sender: TObject);
var
  SelectedItem: TListItem;
begin
  SelectedItem := ListView.Selected;

  if Assigned(SelectedItem) then
    if not BluetoothEnableDiscovery(Integer(SelectedItem.Data),
                  TAction(Sender).Caption = 'Not disc.') then
      MessageDlg('Unable to change Radio state', mtError, [mbOK], 0)

    else
      TreeViewChange(TreeView, TreeView.Selected);
end;

procedure TfmMain.acPropertyUpdate(Sender: TObject);
var
  SelectedNode: TTreeNode;
  SelectedItem: TListItem;
begin
  SelectedNode := TreeView.Selected;
  SelectedItem := ListView.Selected;

  TAction(Sender).Enabled := Assigned(SelectedNode) and
                             Assigned(SelectedItem) and
                             (SelectedNode.ImageIndex > 0);
end;

procedure TfmMain.acPropertyExecute(Sender: TObject);
var
  Info: BLUETOOTH_DEVICE_INFO;
begin
  Info := BLUETOOTH_DEVICE_INFO(ListView.Selected.Data^);
  BluetoothDisplayDeviceProperties(Handle, Info);
end;

procedure TfmMain.btSelectClick(Sender: TObject);
var
  ASelParams: BLUETOOTH_SELECT_DEVICE_PARAMS;
  ASelParamsSize: dword;
begin
  ASelParamsSize := SizeOf(BLUETOOTH_SELECT_DEVICE_PARAMS);
  FillChar(ASelParams, ASelParamsSize, 0);
  with ASelParams do begin
    dwSize := ASelParamsSize;
    hwndParent := Handle;
    fShowRemembered := True;
    fAddNewDeviceWizard := True;
  end;

  BluetoothSelectDevices(@ASelParams);
  BluetoothSelectDevicesFree(@ASelParams);
end;

procedure TfmMain.acEnableUpdate(Sender: TObject);
var
  SelectedNode: TTreeNode;
  SelectedItem: TListItem;
begin
  SelectedNode := TreeView.Selected;
  SelectedItem := ListView.Selected;

  TAction(Sender).Enabled := Assigned(SelectedNode) and
                             Assigned(SelectedItem) and
                             (SelectedNode.ImageIndex = -2);
end;

procedure TfmMain.acEnableExecute(Sender: TObject);
var
  GUID: TGUID;
begin
  GUID := StringToGUID(ListView.Selected.Caption);
  BluetoothSetServiceState(TreeView.Selected.Parent.ImageIndex,
                           BLUETOOTH_DEVICE_INFO(TreeView.Selected.Data^),
                           GUID, 
                           BLUETOOTH_SERVICE_DISABLE);
  TreeViewChange(TreeView, TreeView.Selected);
end;

procedure TfmMain.acRemoveUpdate(Sender: TObject);
begin
  TAction(Sender).Enabled := acProperty.Enabled;
end;

procedure TfmMain.acRemoveExecute(Sender: TObject);
var
  Info: BLUETOOTH_DEVICE_INFO;
  Res: dword;
begin
  Info := BLUETOOTH_DEVICE_INFO(ListView.Selected.Data^);
  Res := BluetoothRemoveDevice(Info.Address);
  if Res <> ERROR_SUCCESS then
    MessageDlg('Device not found', mtError, [mbOK], 0);
  TreeViewChange(TreeView, TreeView.Selected);
end;

procedure TfmMain.acUpdateUpdate(Sender: TObject);
begin
  TAction(Sender).Enabled := acProperty.Enabled;
end;

procedure TfmMain.acUpdateExecute(Sender: TObject);
var
  Info: BLUETOOTH_DEVICE_INFO;
  Res: dword;
  NewName: string;
begin
  if InputQuery(' ', ' ', NewName) then begin
    lstrcpyW(Info.szName, PWideChar(WideString(NewName)));
    Res := BluetoothUpdateDeviceRecord(Info);
    if Res <> ERROR_SUCCESS then RaiseLastOsError;
    TreeViewChange(TreeView, TreeView.Selected);
  end;
end;

end.
